FALSE 	EQU 0
TRUE 	EQU ~FALSE

* Bring smart watch online - will this crash on exit?

SKIP1 EQU $21	BRN
SKIP2 EQU $8C	CMPX #

SPACE	EQU $20
SPACES	EQU $2020

DBASROM EQU $8000
DBASVEC EQU $FFF0

ROMRAMS	EQU $FFDE

PIA1DB	EQU $FF22			; Side B Data/DDR

PIA1CRB	EQU $FF23			; Side B Control.
CRDDRDATA	EQU %00000100	; DDR register accessed if 1, data if 0
MaskROMSEL	EQU %00000100	; ROM select output on Dragon 64

PRTSTR EQU $90E5    ; (Coco B99C)

* Reading a DS1315 style chip internally in a Dragon/CoCo machine has an interesting challenge
* Most instructions on the 6809 emit totally dead cycles, which are a read of address FFFF
* This is decoded by the SAM to be a read of the top byte of the vector area in one of the ROM chips,
* and will be interpreted by the DS1315 as a RTC bit read cycle. When this happens, any key writing
* attempt is aborted, so by default we find that routines to write the 64 bit activation key do not work!

* We need to work around this. On a Dragon 32 we have two 8k ROMs and only one is decoded by the SAM
* as the place to get the 16 bytes of ROM vector values

* Fitting the RTC in the lower (not vector storing) 8k socket, and using the base of this ROM, $8000,
* as the RTC read/write bits base address will ensure we can access the RTC safely in the 32 or 32(64).
* Note that in a 32(64) we might be in all RAM mode, so our access code would have to be in the lower 32k
* as otherwise addressing the ROM will swap out our running program code.

* For a real Dragon 64 we might be in ROM/RAM mode or all RAM mode and we have 2 16k ROMS both of which 
* have vector areas
* Switching to all RAM mode will not take RTC entirely out of picture because the vector area still
* activates the ROM irrespective of which ROM is configured, so dead cycles get seen by the RTC
* The only safe methods are

* 1 Swapping which Dragon64 ROM chip we have selected to the socket with RTC ONLY for the bit reads / writes
* 2 Writing entirely do-not-emit Read FFFF dead cycle read and write routines

* We should be able to use the Vector area on a Dragon 64 as the base address, and this ensures our
* code can work from either RAM/ROM or all RAM mode

* Lets try a routine based on method 1 for the D64 initially.
* Entirely dead cycle safe code could come later but is probably only useful for 32(64) machines with
* a single 32k EPROM board fitted to make them highly 64-like.

* Reads to base address + 0 are writes of a 0 bit
* Reads to base address + 1 are writes of a 1 bit
* Reads to base address with bit 3 (+4) set are reads of the 64 RTC data bits available in sequence
* once we have keyed in successfully (obviously)

* Key sequence is $C5,$3A,$A3,$5C,$C5,$3A,$A3,$5C

START 	ORG $7000
		BRA RSTART		; Enter and do Read Routine
		BRA WRSTART		; Enter and do Write Routine
RSTART	PSHS D,X,Y,U,CC ; Save CC state and registers we mess

		LEAS -3,S		; Make safe access temp space
		BSR SAFESET		; Set up safe access

		LDY #ALLERT		; Get Y pointer to key bit sequence
		LDB #8			; Bytes to write
		
		BSR	RTCWRN		; Write it

		; To read or write to the clock we should be doing it here before restoring to unsafe mode
		; Only restore to unsafe mode once you've completed all RTC access business
		; Otherwise dead cycles will gradually read away many bits you want to access

		; Lets read to test for clock presence
		; If RTC present, reads from the address will return differing values, as bit 0 will be changed by the RTC

		LDY #CLKVALS	; Y pointer space for read bytes
		LDB #8			; Bytes to read
		BSR RTCRN		; Read bytes

		BSR SAFECLR		; Clear down out of safe access mode
		LEAS 3,S		; Reclaim temp stack
		
		LDY #CLKVALS	; Y pointer space for read bytes
		LDB #7			; Bytes to check
		LDA	,Y+
CHKMORE	CMPA ,Y+
		BNE	CFOUND
		DECB
		BNE CHKMORE
		
		; Get here... we have all bytes in the read sequence the same, ie no clock found, report it
		
		LEAX NCLOCK-1,PCR
 		JSR PRTSTR
		
CFOUND	PULS D,X,Y,U,CC,PC ; Exit

WRSTART	PSHS D,X,Y,U,CC ; Save CC state and registers we mess

		LEAS -3,S		; Make safe access temp space
		BSR SAFESET		; Set up safe access

		LDY #ALLERT		; Get Y pointer to key bit sequence
		LDB #8			; Bytes to write
		
		BSR	RTCWRN		; Write it

		; Write the the RTC to set it up

		LDY #CLKVALS	; Y pointer space for write bytes
		LDB #8			; Bytes to write
		BSR RTCWRN		; Write bytes

		BSR SAFECLR		; Clear down out of safe access mode
		LEAS 3,S		; Reclaim temp stack
		
		PULS D,X,Y,U,CC,PC ; Exit

* Need to enable ROM select and switch to the other ROM set here which is safe
* (ie no RTC)
* Ideally the D64 mode ROM would be the one we would have the RTC fitted under
* because that can only be seen with the DDR change and correct output
* but my board only fits under the D32 mode ROM, so we will have to switch to
* D64 mode ROM as 'safe'		
		
SAFESET	LDA PIA1CRB		; get control register of PIA1, port B
        ANDA #~CRDDRDATA ; set flag to access DDR of port b
        STA PIA1CRB		; tell the PIA
		LDB PIA1DB		; get DDR register state
        STB 4,S			; Save DDR
		ORB #MaskROMSEL	; Make ROMSEL bit an output
		ORCC #$50		; Interrupts off now - ROM might go byebye! they'll be restored on exit
		STB PIA1DB		; Configure PIA
		ORA #CRDDRDATA  ; return to accessing data register of port B
        STA PIA1CRB		; tell PIA
		LDB PIA1DB		; get data register of PIA
        TFR B,A         ; copy
        ANDA #~MaskROMSEL ; amend it to select second (64k mode) ROM - safe
        STA PIA1DB      ; update PIA - we are now in safe mode
						; note that just before the write cycle occurs which goes safe, this emits an FFFF Read cycle, making the RTC ready for keying
						
		STD 2,S			; Save the Safe and Unsafe masks
						; 2,S points to the value to write to PIA1DB to go safe
						; 3,S points to the value to write to PIA1DB to go unsafe
		RTS

* Write sequence of N bytes to RTC
* Least significant bit first starting at address in Y
* Bytes to write in B

RTCWRN	PSHS B			; Save count to write
		LDX #DBASVEC	; Point X to Base of Vector area
						; Reads to base address + 4 are actual reads
						; Reads to base address + 0 are writes of a 0 bit
						; Reads to base address + 1 are writes of a 1 bit
		
NXTBYT	LDB	,Y+			; Get the byte
		BSR WRTCBYT		; Write it to RTC
		DEC	,S			; Update byte write count
		BNE NXTBYT		; Keep writing bytes till done
		LEAS 1,S		; Forget stack byte count
		RTS

* Read sequence of N bytes from just-keyed RTC in safe mode
* Least significant bit first writing to at addresses starting at Y
* Bytes to read in B

RTCRN	PSHS B			; Save count to read
		LDX #DBASVEC+4	; Point X to Vector area read address
						; Reads to base address + 4 are actual reads (!WE line on A2)
NXTRBYT	BSR RRTCBYT		; Read byte from RTC
		STB	,Y+			; Save in the data area

		DEC	,S			; Update remaining bytes to read count
		BNE NXTRBYT		; Keep writing bytes till done
		LEAS 1,S		; Forget read byte count
		RTS						
		
* Return to original unsafe ROM select mode
SAFECLR LDD 3,S			; Get unsafe toggle and saved DDR value
		LDY #PIA1DB		; Where to write
		STA ,Y			; Go back to unsafe * warning we are now eating bits on dead cycles, and will trigger a write first during the PULS B *
		LDA PIA1CRB		; get control register of PIA1, port B
        ANDA #~CRDDRDATA ; set flag to access DDR of port b
        STA PIA1CRB		; tell the PIA
		STB PIA1DB		; restore DDR
		ORA #CRDDRDATA  ; return to accessing data register of port B
        STA PIA1CRB		; tell PIA
		
		RTS		
	
* write byte in B
* Entry from 64k mode ROM (safe) toggles in to 32K mode ROM and back out again to achieve each safe write of a bit to RTC
* X points to base of ROM / RTC

WRTCBYT	PSHS B			; Byte we are writing
		LDA #8			; How many bits to write
		PSHS A			; Save it
		LEAU 1,X		; Get RTC write address for setting a bit 
		PSHS Y			; Save current byte in sequence pointer
		LDY #PIA1DB		; Get pointer to PIA data reg to update for safe mode switching
		LDA 9,S			; Get safe on and off values in A and B respectively
NXBIT	LDB 10,S
		LSR 3,S			; Shift the bits in the byte we are writing
		BCS HIGH		; And go to write it
		STB	,Y			; enter (unsafe) D32 mode ROM
		LDB ,X 			; write clear bit to DS1315 (no dead cycle emitted)
		STA ,Y 			; Go back to Safe 64k ROM mode (doesn't emit a dead cycle first)
		BRA SKIP
HIGH	STB	,Y			; enter (unsafe) D32 mode ROM
		LDB ,U 			; write set bit to DS1315 (no dead cycle emitted)
		STA ,Y 			; Go back to Safe 64k ROM mode (doesn't emit a dead cycle first)
SKIP	DEC 2,S
		BNE NXBIT
		PULS Y			; restore byte sequence pointer in Y
		LEAS 2,S		; reframe stack discarding shifting / write byte 
		RTS
						
* Read byte into B from RTC
* Entry with 64k mode ROM (safe) toggles in to 32K mode ROM, reads 8 bits then comes out again

RRTCBYT	PSHS Y			; Save current byte in sequence pointer	
		LDY #PIA1DB		; Get pointer to PIA data reg to update for safe mode switching
		LDD 7,S			; Get safe on and off values in A and B respectively
		PSHS A			; Put return to safe value in A somewhere safely accessible
		
		STB	,Y			; enter (unsafe) D32 mode ROM
		
		LDA ,X 			; get bit 0 from DS1315 (3 cycles, safe)
		LSRA			; Move it to C (2 cycles, safe)
		RORB			; Move it in to read byte (2 cycles, safe)

		LDA ,X 			; and for bit 1
		LSRA
		RORB			
		LDA ,X 			; 2
		LSRA
		RORB			
		LDA ,X 			; 3
		LSRA
		RORB			
		LDA ,X 			; 4
		LSRA
		RORB			
		LDA ,X 			; 5
		LSRA
		RORB			
		LDA ,X 			; 6
		LSRA
		RORB			
		LDA ,X 			; 7
		LSRA
		RORB			

		LDA ,S			; Get safe mode mask again (3 cyles, safe)
		STA ,Y 			; Go back to Safe 64k ROM mode (3 cycles, safe)
		PULS A,Y,PC
		
CLKVALS	RMB	8

ALLERT 	FCB $C5,$3A,$A3,$5C,$C5,$3A,$A3,$5C,0

		FCB 0
NCLOCK	FCC /CLOCK NOT FOUND/
 		FCB 0



