********************************************************************
* getrtc - Get real time clock time
*
* OS9 module to read DS1315 chip inside D64 in ROM1 (32k mode ROM) slot
* uses approach swapping ROM1 out of use (for ROM2) between bit reads and writes
* because we cannot allow dead cycles to appear at the ROM1, they mess up 
* our read / write sequences

* Program emits 64 bit key sequence, then reads 8 bytes into local storage
* this is then transformed into a string and written out to stdio
* default format is compatible with os9 settime input
* yy mm dd hh mm ss

* other formats may be implemented later with command line switches, suggest
* suggestions
* -d date only
* -t time only
* -n nitros9 settime yyyy mm dd hh mm ss format

* companion program to set rtc from hex sequence or date string?

*
* Edt/Rev  YYYY/MM/DD  Modified by
* Comment
* ------------------------------------------------------------------
*   1      2021/11/22  mike miller
* Started building.

         nam   getrtc
         ttl   Get Dragon64 ROM1 DS1315 rtc time

 ifp1
          use   defsfile
 endc

* Vector area & ROM access control PIA address and mask values

DBASVEC EQU $FFF0
PIA1DB	EQU $FF22			; Side B Data/DDR
PIA1CRB	EQU $FF23			; Side B Control.
CRDDRDATA	EQU %00000100	; DDR register accessed if 1, data if 0
MaskROMSEL	EQU %00000100	; ROM select output on Dragon 64

* Here are some tweakable options
DOHELP   set   0    
STACKSZ  set   32    estimated stack size in bytes
PARMSZ   set   256   estimated parameter size in bytes

* Module header definitions
tylg     set   Prgrm+Objct   
atrv     set   ReEnt+rev
rev      set   $01
edition  set   1

         mod   eom,name,tylg,atrv,start,size

         org   0
clkvals  rmb  8
dtstring rmb  64
		 
cleartop equ   .    everything up to here gets cleared at start

* Finally the stack for any PSHS/PULS/BSR/LBSRs that we might do
         rmb   STACKSZ+PARMSZ
size     equ   .

* The utility name and edition goes here
name     fcs   /getrtc/
         fcb   edition

* Place constant strings here

ALLERT	 fcb $C5,$3A,$A3,$5C,$C5,$3A,$A3,$5C,0

dfmask   fcb $8F,$1F,$81,$1F,$83,$4F,$00,$83,$2F,$87,$2F,$87,$0F,$00

RTCnf fcc /RTC not found!/
CR        fcb C$LF
          fcb C$CR
RTCnfL equ   *-RTCnf

* Here's how registers are set when this process is forked:
*
*   +-----------------+  <--  Y          (highest address)
*   !   Parameter     !
*   !     Area        !
*   +-----------------+  <-- X, SP
*   !   Data Area     !
*   !   & Direct Page !
*   +-----------------+  <-- U, DP       (lowest address)
*
*   D = parameter area size
*  PC = module entry point abs. address
*  CC = F=0, I=0, others undefined

* The start of the program is here.

* main program
start		
			
		pshs u,cc		; Save u	
		leas -3,s		; Make safe access temp space
		lbsr safeset		; Set up safe access

		leay ALLERT,pcr	; Get Y pointer to key bit sequence
		ldb #8			; Bytes to write
		
		lbsr	rtcwrn		; Write it

		; To read or write to the clock we should be doing it here before restoring to unsafe mode
		; Only restore to unsafe mode once you've completed all RTC access business
		; Otherwise dead cycles will gradually read away many bits you want to access

		; Lets read to test for clock presence
		; If RTC present, reads from the address will return differing values, as bit 0 will be changed by the RTC
		
		ldy 4,s			; get stacked U value into Y (pointer space for read bytes clkvals)
		ldb #8			; Bytes to read
		lbsr rtcrn		; Read bytes

		lbsr safeclr		; Clear down out of safe access mode
		leas 3,S		; Reclaim temp stack
		puls u,cc		; Restore u and cc (re-enable interrupts)
		
		leay ,u			; setup to check clock data is valid
		ldb #7			; Bytes to check
		lda	,y+			; first byte
chkmore	cmpa ,y+		; compare with next
		bne cfound		; not the same, all good
		decb			; otherwise, the same, we have one less to check
		bne chkmore		; check next
		
* Get here... we have all bytes in the read sequence the same, ie no clock found, report it
				
noRTCEr leax RTCnf,pcr
		ldy #RTCnfL

ExitEr	ldd #$0201 		; std error and throw error number 1
ExitPr	os9 I$Write
Exitnow	os9 F$Exit

* Now we have grabbed the RTC bytes and they appear valid, lets
* build a human readable representation of the RTC date / time data for output

* Do the other bits

cfound	
		leay clkvals+7,u	start of date time bytes at year end
		leax dfmask,pcr		byte masks pointer
		ldb #8				done counter
		pshs b,u
		leau dtstring,u		string out pointer
nxnyb	lda ,y				get clock byte
		ldb ,x				get mask / instruction byte
		bpl lnib			negative bit indicates this is a high nibble
		lsra				shift upper
		lsra				nybble
		lsra				into
		lsra				lower
lnib	anda ,x+			mask out anything we don't want
		anda #$0F			including any 'instruction' bits
		tstb				check mask / instruction byte agaaaain
		beq	skipo 			if mask byte is zero skip nybble output

		adda #'0			convert to ascii number
		sta ,u+
skipo	tstb				check mask / instruction byte again
		bmi nxnyb			next is lower half of same input byte
		leay -1,y			move to next input byte
		dec ,s			    reduce byte count to do
		beq done
		tstb
		beq nxnyb 
		bitb #$10			is the / separator flag set
		bne fsep			setup for forward slash character separator
		bitb #$20			is the : separator flag set
		bne csep			setup for colon character separator
		bitb #$40
		bne spsep
		bra nxnyb			no separator		
outsep	stb ,u+				put in output string
		bra nxnyb			do the next nybble
fsep	ldb #'/				forward
		bra outsep			output it
csep    ldb #':				colon
		bra outsep			output it
spsep   ldb #$20			setup for space character		
		bra outsep			output it

* Hopefully here we have a nice output string built for us		
* (must be a neater way to do this, bleurgh)
done	leas 1,s			forget about b
		clr ,u+
		tfr u,d
		puls u
		leax dtstring,u
		pshs x
		subd ,s
		tfr d,y
		
		ldd #$0100		; std out and no error
		bra  ExitPr  	; Exit with print to std out

* Need to enable ROM select and switch to the other ROM set here which is safe
* (ie no RTC)
* Ideally the D64 mode ROM would be the one we would have the RTC fitted under
* because that can only be seen with the DDR change and correct output
* but my board only fits under the D32 mode ROM, so we will have to switch to
* D64 mode ROM as 'safe'		
		
safeset	lda PIA1CRB		; get control register of PIA1, port B
        anda #~CRDDRDATA ; set flag to access DDR of port b
        sta PIA1CRB		; tell the PIA
		ldb PIA1DB		; get DDR register state
        stb 4,S			; Save DDR
		orb #MaskROMSEL	; Make ROMSEL bit an output
		orcc #$50		; Interrupts off now - ROM might go byebye! they'll be restored on exit
		stb PIA1DB		; Configure PIA
		ora #CRDDRDATA  ; return to accessing data register of port B
        sta PIA1CRB		; tell PIA
		ldb PIA1DB		; get data register of PIA
        tfr B,A         ; copy
        anda #~MaskROMSEL ; amend it to select second (64k mode) ROM - safe
        sta PIA1DB      ; update PIA - we are now in safe mode
						; note that just before the write cycle occurs which goes safe, this emits an FFFF Read cycle, making the RTC ready for keying
						
		std 2,S			; Save the Safe and Unsafe masks
						; 2,S points to the value to write to PIA1DB to go safe
						; 3,S points to the value to write to PIA1DB to go unsafe
		rts

* Write sequence of N bytes to RTC
* Least significant bit first starting at address in Y
* Bytes to write in B

rtcwrn	PSHS b			; Save count to write
		ldx #DBASVEC	; Point X to Base of Vector area
						; Reads to base address + 4 are actual reads
						; Reads to base address + 0 are writes of a 0 bit
						; Reads to base address + 1 are writes of a 1 bit
		
nxtbyt	ldb	,y+			; Get the byte
		bsr wrtcbyt		; Write it to RTC
		dec	,s			; Update byte write count
		bne nxtbyt		; Keep writing bytes till done
		leas 1,s		; Forget stack byte count
		rts

* Read sequence of N bytes from just-keyed RTC in safe mode
* Least significant bit first writing to at addresses starting at Y
* Bytes to read in B

rtcrn	pshs b			; Save count to read
		ldx #DBASVEC+4	; Point X to Vector area read address
						; Reads to base address + 4 are actual reads (!WE line on A2)
nxtrbyt	bsr rrtcbyt		; Read byte from RTC
		stb	,y+			; Save in the data area

		dec	,s			; Update remaining bytes to read count
		bne nxtrbyt		; Keep writing bytes till done
		leas 1,s		; Forget read byte count
		rts						
		
* Return to original unsafe ROM select mode
safeclr ldd 3,s			; Get unsafe toggle and saved DDR value
		ldy #PIA1DB		; Where to write
		sta ,Y			; Go back to unsafe * warning we are now eating bits on dead cycles, and will trigger a write first during the PULS B *
		lda PIA1CRB		; get control register of PIA1, port B
        anda #~CRDDRDATA ; set flag to access DDR of port b
        sta PIA1CRB		; tell the PIA
		stb PIA1DB		; restore DDR
		ora #CRDDRDATA  ; return to accessing data register of port B
        sta PIA1CRB		; tell PIA
				
		rts
		
* write byte in B
* Entry from 64k mode ROM (safe) toggles in to 32K mode ROM and back out again to achieve each safe write of a bit to RTC
* X points to base of ROM / RTC

wrtcbyt	pshs b			; Byte we are writing
		lda #8			; How many bits to write
		pshs a			; Save it
		leau 1,x		; Get RTC write address for setting a bit 
		pshs y			; Save current byte in sequence pointer
		ldy #PIA1DB		; Get pointer to PIA data reg to update for safe mode switching
		lda 9,s			; Get safe on and off values in A and B respectively
nxbit	ldb 10,s
		lsr 3,s			; Shift the bits in the byte we are writing
		bcs high		; And go to write it
		stb	,y			; enter (unsafe) D32 mode ROM
		ldb ,x 			; write clear bit to DS1315 (no dead cycle emitted)
		sta ,y 			; Go back to Safe 64k ROM mode (doesn't emit a dead cycle first)
		bra skip
high	stb	,y			; enter (unsafe) D32 mode ROM
		ldb ,u 			; write set bit to DS1315 (no dead cycle emitted)
		sta ,y 			; Go back to Safe 64k ROM mode (doesn't emit a dead cycle first)
skip	dec 2,s
		bne nxbit
		puls y			; restore byte sequence pointer in Y
		leas 2,s		; reframe stack discarding shifting / write byte 
		rts
						
* Read byte into B from RTC
* Entry with 64k mode ROM (safe) toggles in to 32K mode ROM, reads 8 bits then comes out again

rrtcbyt	pshs y			; Save current byte in sequence pointer	
		ldy #PIA1DB		; Get pointer to PIA data reg to update for safe mode switching
		ldd 7,s			; Get safe on and off values in A and B respectively
		pshs a			; Put return to safe value in A somewhere safely accessible
		
		stb	,y			; enter (unsafe) D32 mode ROM
		
		lda ,x 			; get bit 0 from DS1315 (3 cycles, safe)
		lsra			; Move it to C (2 cycles, safe)
		RORB			; Move it in to read byte (2 cycles, safe)

		lda ,x 			; and for bit 1
		lsra
		rorb			
		lda ,x 			; 2
		lsra
		rorb			
		lda ,x 			; 3
		lsra
		rorb			
		lda ,x 			; 4
		lsra
		rorb			
		lda ,x 			; 5
		lsra
		rorb			
		lda ,x 			; 6
		lsra
		rorb			
		lda ,x 			; 7
		lsra
		rorb			

		lda ,s			; Get safe mode mask again (3 cyles, safe)
		sta ,y 			; Go back to Safe 64k ROM mode (3 cycles, safe)
		puls a,y,pc

        emod
eom     equ   *
        end


